let sqlite3 = require('sqlite3');
let bcrypt = require('bcrypt-nodejs');
let TokenStore = require('passwordless-tokenstore');

const makeDbHandle = (fileName, mode) => {
    fileName = fileName || '';
    mode = mode || (sqlite3.OPEN_READWRITE | sqlite3.OPEN_CREATE);
    let dbHandle = new sqlite3.Database(fileName, mode, (err) => {
        if (err) {
            throw new Error('Sqlite3 error: Could not instantiate handle. Possible cause: ' + err);
        }
    });
    return dbHandle;
};

const tableName = new WeakMap();
const fileName = new WeakMap();
const accessMode = new WeakMap();
const dbConn = new WeakMap();

class Sqlite3Store extends TokenStore {

    constructor (sqlFile, options) {
        super();
        options = options || {};

        let mode = sqlite3.OPEN_READWRITE | sqlite3.OPEN_CREATE;
        if (typeof options.mode !== 'undefined') { mode = options.mode; }
        accessMode.set(this, mode);

        let table = 'passwordless';
        if (typeof options.table !== 'undefined') { table = options.table; }
        tableName.set(this, table);

        fileName.set(this, sqlFile || "");

        try {
            let conn = makeDbHandle(fileName.get(this), accessMode.get(this));
            dbConn.set(this, conn);
            this.setUpPasswordless();
        } catch (err) {
            console.log(err);
        }
    }

    setUpPasswordless () {
        const dbHandle = dbConn.get(this);
        const passwordlessTable = tableName.get(this);
        const sqlQuery = "CREATE TABLE IF NOT EXISTS '" + passwordlessTable + "' ( `uid` TEXT NOT NULL, `token`  TEXT NOT NULL, `originUrl`	TEXT, `ttl`	INTEGER, PRIMARY KEY(`uid`) );";
        dbHandle.run(sqlQuery);
    }

    authenticate (token, uid, callback) {
        if(!token || !uid || !callback) {
            throw new Error('Sqlite3Store:authenticate called with invalid parameters');
        }

        const dbHandle = dbConn.get(this);
        const table = tableName.get(this);
        const sqlQuery = "SELECT * FROM " + table + " WHERE uid='" + uid + "'";
        dbHandle.all(sqlQuery, (err, rows) => {
            if (err) {
                return callback(err, false, null);
            } else if(!rows || !rows.length || (rows && rows.length > 1)) {
                return callback(null, false, null);
            } else if(Date.now() > rows[0].ttl) {
                callback(null, false, null);
            } else {
                if (bcrypt.compareSync(token, rows[0].token)) {
                    if (rows[0].originUrl !== 'null') {
                        callback(null, true, rows[0].originUrl);
                    } else {
                        callback(null, true, "");
                    }
                } else {
                    callback(null, false, null);
                }
            }
        });
    }

    storeOrUpdate (token, uid, msToLive, originUrl, callback) {
        if(!token || !uid || !msToLive || !callback) {
            throw new Error('Sqlite3Store:storeOrUpdate called with invalid parameters');
        }

        const dbHandle = dbConn.get(this);
        const table = tableName.get(this);
        const ttl = Date.now() + msToLive;
        const hash = bcrypt.hashSync(token);
        const sqlQuery = "INSERT OR REPLACE INTO " + table + "(uid, token, originUrl, ttl) VALUES ('" + uid +"', '" + hash + "', '" + originUrl + "', "+ ttl +" );";
        dbHandle.run(sqlQuery, (err) => {
            if (err) {
                this.length(err, (rowCount) => { console.log("rowCount: " + rowCount)});
                callback(err);
            } else {
                this.length(err, (rowCount) => { console.log("rowCount: " + rowCount)});
                callback();
            }
        });

    }

    invalidateUser (uid, callback) {
        if (!uid || !callback) {
            throw new Error('Sqlite3Store:invalidateUser called with invalid parameters');
        }
        const dbHandle = dbConn.get(this);
        const table = tableName.get(this);
        const sqlQuery = "DELETE FROM " + table + " WHERE uid='" + uid + "';";
        dbHandle.exec(sqlQuery, (err) => {
            if (err) {
                callback(err);
            } else {
                callback();
            }
        });
    }

    clear (callback) {
        if (!callback) {
            throw new Error('Sqlite3Store:clear called with invalid parameters');
        }
        const dbHandle = dbConn.get(this);
        const table = tableName.get(this);
        const sqlQuery = "DELETE FROM " + table + ";";
        dbHandle.exec(sqlQuery, (err) => {
            if (callback && typeof callback === 'function') {
                if (err) {
                    callback(err);
                } else {
                    callback();
                }
            }
        });
    }

    length (callback) {
        const dbHandle = dbConn.get(this);
        const table = tableName.get(this);
        const sqlQuery = "SELECT count(*) FROM " + table + ";";
        dbHandle.all(sqlQuery, (err, rows) => {
            if (callback && typeof callback === 'function') {
                if (err) {
                    callback(err);
                } else {
                    callback(null, parseInt(rows[0]['count(*)'], 10));
                }
            }
        })
    }
}

module.exports = Sqlite3Store;